/*
 * SetRes -- A general-purpose Widget resource testing module
 *
 * HEADER
 *    #include "SetRes.h"
 *
 * PROTOTYPE
 *    void SetRes(Widget widget)
 *
 * ARGUMENTS
 *    Widget widget
 *       Specifies any widget descendant of the shell for which you
 *       want to create a SetRes dialog.  SetRes will not be able to
 *       identify a widget that is not also a descendant of this shell.
 *      
 * DESCRIPTION
 *    The SetRes dialog allows you to set the resources on any widget in
 *    an application.  The dialog contains two text fields; the first
 *    specifies the widget in question and the second should be set to
 *    a value of the form resource_name:resource_value.  In place of a
 *    resource name/value pair, the special commands "unmanage", "manage",
 *    and "destroy" can be used to unmanage, manage, or destroy the named
 *    widget.
 *   
 * USAGE
 *    The widget may be specified with the full syntax supported by
 *    XtNameToWidget(), including '*' and '.' characters.  You will
 *    generally need to precede widget names with a '*' or provide
 *    qualified names beginning with direct descendants of the shell.
 *
 *    The resource field may also be entered in a form like
 *    'resource_name:!<n>', where <n> is any integer, to explicitly set
 *    the resource value to <n>.  This can be useful when specifying values
 *    for enumerated resource types for which a string-to-type converter
 *    is not available.
 * 
 * AUTHOR
 *    Alvin Wen
 */ 
#include <stdio.h>
#include <ctype.h>
#include <X11/Intrinsic.h>
#include <Xm/Form.h>
#include <Xm/DialogS.h>
#include <Xm/TextF.h>
#include <Xm/LabelG.h>
#include <Xm/PushBG.h>
#include <Xm/SeparatoG.h>
#include "SetRes.h"

extern int atoi(
#if NeedFunctionPrototypes
    char *
#endif
);

static Widget Shell, Setres=NULL;

typedef struct {
    String name, type;
    Cardinal size;
} resource_t;

typedef struct {
    Widget name, value;
} cdata_t;

static resource_t *Resources;
static int numResources;

/*
 * The string-to-widget converter
 */
#if NeedFunctionPrototypes
Boolean CvtStringToWidget(Display *dpy,
			    XrmValuePtr args, Cardinal *num_args,
			    XrmValuePtr from, XrmValuePtr to,
			    XtPointer *data)
#else
Boolean CvtStringToWidget(dpy, args, num_args, from, to, data)
Display *dpy;
XrmValuePtr args; Cardinal *num_args;
XrmValuePtr from; XrmValuePtr to;
XtPointer *data;
#endif
{
    static Widget widget;
    Boolean success = True;

    if ((widget=XtNameToWidget(Shell, from->addr))==NULL) {
	XtDisplayStringConversionWarning(dpy, from->addr, "Widget");
	return False;
    }
    if (to->addr==NULL)
	to->addr = (XtPointer)&widget;
    else if (to->size < sizeof(Widget))
	success = False;
    else
	*(Widget *)to->addr = widget;
    to->size = sizeof(Widget);		
    return success;
}

/*
 * Collect a widget's normal and constraint resources in the single
 * conveniently accessible structure
 */
#if NeedFunctionPrototypes
static void CopyResources(Widget widget)
#else
static void CopyResources(widget)
Widget widget;
#endif
{
    int i;
    Cardinal num_resources, num_consts;
    XtResourceList resources, consts;

    XtGetResourceList(XtClass(widget), &resources, &num_resources);
    XtGetConstraintResourceList(XtClass(widget), &consts, &num_consts);
    if (Resources)
	XtFree((char *)Resources);
    numResources = num_resources + num_consts;
    Resources = (resource_t *)XtMalloc(numResources*sizeof(resource_t));
    for (i=0; i < num_resources; i++) {
	Resources[i].name = XtNewString(resources[i].resource_name);
	Resources[i].type = XtNewString(resources[i].resource_type);
	Resources[i].size = resources[i].resource_size;
    }
    XtFree((char *)resources);
    for (i=0; i < num_consts; i++) {
	Resources[num_resources+i].name = XtNewString(consts[i].resource_name);
	Resources[num_resources+i].type = XtNewString(consts[i].resource_type);
	Resources[num_resources+i].size = consts[i].resource_size;
    }
    XtFree((char *)consts);
}

/*
 * Free the resource structure
 */
#if NeedFunctionPrototypes
static void DestroyResources(void)
#else
static void DestroyResources()
#endif
{
    int i;
    if (Resources) {
	for (i=0; i < numResources; i++) {
	    XtFree((char *)Resources[i].name);
	    XtFree((char *)Resources[i].type);
	}
	XtFree((char *)Resources);
    }
    Resources = NULL;
    numResources = 0;
}

/*
 * Separate a resource_name:resource_value pair into its two components.
 * If no resource value is specified, value[0] will be returned NUL
 */
#if NeedFunctionPrototypes
static void ParseResource(String resource_string, char *name, char *value)
#else
static void ParseResource(resource_string, name, value)
String resource_string;
char *name; char *value;
#endif
{
    int i;
    while (isspace(*resource_string))
	resource_string++;
    i=0;
    while (*resource_string && *resource_string != ':')
	name[i++] = *resource_string++;
    name[i] = '\0';

    if (*resource_string==':')
	resource_string++;

    while (isspace(*resource_string))
	resource_string++;

    i=0;
    while (*resource_string)
	value[i++] = *resource_string++;
    value[i] = '\0';
}
	    
/*
 * Parse, convert, and set the resource on the named widget
 */
#if NeedFunctionPrototypes
static void ApplyCB(Widget button, XtPointer client, XtPointer cbs)
#else
static void ApplyCB(button, client, cbs)
Widget button; XtPointer client; XtPointer cbs;
#endif
{
    Arg arg;
    int i, val;
    Widget widget;
    XrmValue from, to;
    static Widget previous;
    cdata_t *cdata = (cdata_t *)client;
    String widget_name, resource_string;
    char resource_name[256], resource_value[256];

    XtVaGetValues(cdata->name, XmNvalue, &widget_name, NULL);
    XtVaGetValues(cdata->value, XmNvalue, &resource_string, NULL);

    if ((widget=XtNameToWidget(Shell, widget_name))==NULL) {
	char buf[256];
	sprintf(buf, "SetRes: Couldn't find widget named '%s'", widget_name);
	XtWarning(buf);
	return;
    }

    if (previous != widget) {
	previous = widget;
	DestroyResources();
	CopyResources(widget);
    }

    if (!resource_string)
	return;

    ParseResource(resource_string, resource_name, resource_value);
    if (*resource_value=='\0') {
	if (!strcmp(resource_name, "manage")) {
	    XtManageChild(widget);
	    return;
	}
	else if (!strcmp(resource_name, "unmanage")) {
	    XtUnmanageChild(widget);
	    return;
	}
	else if (!strcmp(resource_name, "destroy")) {
	    XtDestroyWidget(widget);
	    DestroyResources();
	    previous = NULL;
	    return;
	}
	else {
	    char buf[256];
	    sprintf(buf, "SetRes: Unrecognized command '%s'", resource_name);
	    XtWarning(buf);
	    return;
	}
    }
	
    for (i=0; i < numResources; i++)
	if (!strcmp(Resources[i].name, resource_name)) {
	    break;
	}
    if (i==numResources) {
	char buf[256];
	sprintf(buf, "SetRes: Couldn't find resource named '%s'",
		resource_name);
	XtWarning(buf);
	return;
    }
    if (*resource_value=='!' && Resources[i].size <= 4) {
	/* User explicitly wants to set resource value to specified number */
	val = atoi(resource_value+1);
	to.addr = (XPointer)&val;
	to.size = Resources[i].size;
    }
    else { 
	from.addr = resource_value;
	from.size = 1+strlen(resource_value);
	to.addr = NULL;
	if (!XtConvertAndStore(widget, XtRString, &from,
						Resources[i].type, &to))
	return;
    }
    arg.name = Resources[i].name;
    if (to.size > 4 || !strcmp(Resources[i].type, XtRString))
	arg.value = (XtArgVal)to.addr;
    else switch (to.size) {
	case 1:
	    arg.value = *(char *)to.addr;
	    break;
	case 2:
	    arg.value = *(unsigned short *)to.addr;
	    break;
	case 4:
	    arg.value = *(unsigned int *)to.addr;
	    break;
    }
    XtSetValues(widget, &arg, 1);
}

/*
 * Destroy the SetRes dialog and free the resource structure
 */
#if NeedFunctionPrototypes
static void DismissCB(Widget button, XtPointer client, XtPointer cbs)
#else
static void DismissCB(button, client, cbs)
Widget button; XtPointer client; XtPointer cbs;
#endif
{
    XtDestroyWidget(Setres);
    DestroyResources();
    Setres = NULL;
}

/*
 * Instantiate the SetRes dialog for descendants of the WMShell parent
 * of the widget passed
 */
#if NeedFunctionPrototypes
void SetRes(Widget widget)
#else
void SetRes(widget)
Widget widget;
#endif
{
    static cdata_t cdata;
    XmString label_string;
    Widget form, name, value, nameLabel, valueLabel, sep, apply, dismiss;

    for (Shell = widget; Shell && !XtIsWMShell(Shell);
	    Shell = XtParent(Shell));

    if (Shell==NULL) {
	XtError("SetRes: Couldn't find top level shell widget\n");
	return;
    }

    if (Setres==NULL) {
	XtSetTypeConverter(XtRString, XtRWidget, CvtStringToWidget, NULL, 0,
			    XtCacheAll, NULL);

	Setres = XmCreateDialogShell(Shell, "setres", NULL, 0);
	form = XtVaCreateWidget("form", xmFormWidgetClass, Setres,
				XmNwidth, 350,
				XmNfractionBase, 100,
				XmNautoUnmanage, False,
				NULL);
	label_string = XmStringCreateSimple("Widget:");
	nameLabel = XtVaCreateManagedWidget("nameLabel", xmLabelGadgetClass,
					form,
					XmNlabelString, label_string,
					XmNalignment, XmALIGNMENT_END,
					XmNtopAttachment, XmATTACH_FORM,
					XmNleftAttachment, XmATTACH_FORM,
					XmNrightAttachment, XmATTACH_POSITION,
					XmNrightPosition, 20,
					NULL);
	XmStringFree(label_string);
	name = XtVaCreateManagedWidget("name", xmTextFieldWidgetClass, form,
					XmNcolumns, 25,
					XmNtopAttachment, XmATTACH_FORM,
					XmNrightAttachment, XmATTACH_FORM,
					XmNleftAttachment, XmATTACH_WIDGET,
					XmNleftWidget, nameLabel,
					NULL);
	label_string = XmStringCreateSimple("Resource:");
	valueLabel = XtVaCreateManagedWidget("valueLabel", xmLabelGadgetClass,
					form,
					XmNlabelString, label_string,
					XmNalignment, XmALIGNMENT_END,
					XmNtopAttachment, XmATTACH_WIDGET,
					XmNtopWidget, name,
					XmNleftAttachment, XmATTACH_FORM,
					XmNrightAttachment, XmATTACH_POSITION,
					XmNrightPosition, 20,
					NULL);
	XmStringFree(label_string);
	value = XtVaCreateManagedWidget("value", xmTextFieldWidgetClass, form,
					XmNcolumns, 25,
					XmNtopAttachment, XmATTACH_WIDGET,
					XmNtopWidget, name,
					XmNrightAttachment, XmATTACH_FORM,
					XmNleftAttachment, XmATTACH_WIDGET,
					XmNleftWidget, valueLabel,
					NULL);
	sep = XtVaCreateManagedWidget("separator", xmSeparatorGadgetClass,
					form,
					XmNtopAttachment, XmATTACH_WIDGET,
					XmNtopWidget, value,
					XmNleftAttachment, XmATTACH_FORM,
					XmNrightAttachment, XmATTACH_FORM,
					NULL);
	label_string = XmStringCreateSimple("Apply");
	apply = XtVaCreateManagedWidget("apply", xmPushButtonGadgetClass, form,
					XmNlabelString, label_string,
					XmNtopAttachment, XmATTACH_WIDGET,
					XmNtopWidget, sep,
					XmNleftAttachment, XmATTACH_POSITION,
					XmNleftPosition, 15,
					XmNrightAttachment, XmATTACH_POSITION,
					XmNrightPosition, 45,
					NULL);
	XmStringFree(label_string);
	label_string = XmStringCreateSimple("Dismiss");
	dismiss = XtVaCreateManagedWidget("dismiss", xmPushButtonGadgetClass,
					form,
					XmNlabelString, label_string,
					XmNtopAttachment, XmATTACH_WIDGET,
					XmNtopWidget, sep,
					XmNleftAttachment, XmATTACH_POSITION,
					XmNleftPosition, 55,
					XmNrightAttachment, XmATTACH_POSITION,
					XmNrightPosition, 85,
					NULL);
	XmStringFree(label_string);
	XtVaSetValues(form, XmNdefaultButton, apply, NULL);
	cdata.name = name;
	cdata.value = value;
	XtAddCallback(apply, XmNactivateCallback, ApplyCB, (XtPointer)&cdata);
	XtAddCallback(dismiss, XmNactivateCallback, DismissCB, NULL);
    }
    XtManageChild(form);
}
