#include <Xmt/Xmt.h>
#include <Xmt/Dialog.h>
#include <Xmt/Create.h>
#include <Xmt/Converters.h>
#include <Xmt/Procedures.h>

#include <Xm/List.h>

#define PARITY_EVEN   0
#define PARITY_ODD    1
#define PARITY_MARK   2
#define PARITY_SPACE  3
#define PARITY_NONE   4

#define OPTION_ECHO         0x01
#define OPTION_FLOW_CONTROL 0x02
#define OPTION_ADD_CR       0x04
#define OPTION_SWAP         0x08
#define OPTION_8BIT         0x10

typedef struct {
    int baud;
    int parity;
    int databits;
    int stopbits;
    int flags;
} CommunicationParameters;

XtResource parameter_resources[] = {
{"baud", "baud", XtRInt,
     sizeof(int), XtOffsetOf(CommunicationParameters, baud),
     XtRImmediate, (XtPointer) 9600},
{"parity", "parity", XtRInt,
     sizeof(int), XtOffsetOf(CommunicationParameters, parity),
     XtRImmediate, (XtPointer) PARITY_EVEN},
{"databits", "databits", XtRInt,
     sizeof(int), XtOffsetOf(CommunicationParameters, databits),
     XtRImmediate, (XtPointer) 8},
{"stopbits", "stopbits", XtRInt,
     sizeof(int), XtOffsetOf(CommunicationParameters, stopbits),
     XtRImmediate, (XtPointer) 1},
{"flags", "flags", XtRInt,
     sizeof(int), XtOffsetOf(CommunicationParameters, flags),
     XtRImmediate, (XtPointer) (OPTION_FLOW_CONTROL | OPTION_8BIT)},
};

/* the Directory Entry protocol types for the type field */
#define PROTOCOL_MODEM   0
#define PROTOCOL_TELNET  1
#define PROTOCOL_WWW     2
#define PROTOCOL_GOPHER  3
#define PROTOCOL_FTP     4
#define PROTOCOL_WAIS    5

typedef struct {
    char name[50];
    char address[50];
    char *comments;
    int type;
    CommunicationParameters params;
} DirectoryEntry;

DirectoryEntry directory[] = {
{"Global Network Navigator Home Page",
     "http://gnn.ora.com/foo/bar",
     "GNN from O'Reilly & Associates is one of the coolest\n"
     "WWW nodes out there. It is an online magazine, an\n"
     "internet directory, and an online advertising center.\n"
     "A lot of slick graphic design has gone into it.\n"
     "\n"
     "Look here for pointers to more net resources.",
     PROTOCOL_WWW,
     {9600,PARITY_EVEN,8,1,0}},
{"Electronic Coven",
     "coven.sister.org",
     "Gopher server for inspirational and organizational\n"
     "information of interest to eco-feminists, witches,\n"
     "and others working on the worldly, spiritual, and\n"
     "electronic planes to strengthen the sisterhood and\n"
     "dissolve the patriarchy.  (Not yet implemented.)",
     PROTOCOL_GOPHER,
     {9600,PARITY_EVEN,8,1,0}},
{"Pirate's Cove",
     "+1 (617) 555 6666",
     "A fictional BBS serving the software pirate and hacker.\n"
     "community.  You'll find tons of useful information here\n"
     "on cracking passwords, creating PC viruses, and other\n"
     "guerilla techniques to undermine the information\n"
     "technocracy.",
     PROTOCOL_MODEM,
     {9600,PARITY_EVEN,8,1,0}},
{"Meat-eaters Anonymous",
     "http://death.slaughterhouse.org/ma-info.html",
     "An innovative electronic 12-step program designed for\n"
     "individuals addicted to meat.  Emphasis is placed on\n"
     "gaining awareness of negative ecological, economic, and\n"
     "ethical consequences associated with the consumption of\n"
     "animal flesh, and on establishing a positive relationship\n"
     "with the earth.",
     PROTOCOL_WWW,
},
{"Rooftop Gardeners' Network"},
{"Biodiversity Forum"},
{"Dovetail Systems' System"},
{"Boston Area Weather"},
{"WUMB Program Guide"},
{"Late Night Food @ MIT"},
{"New School for Social Research"},
{"U. of Georgia"},
{"Bastyr College"},
};


XtResource directory_resources[] = {
{"name", "name", XmtRBuffer,
     XmtSizeOf(DirectoryEntry, name), XtOffsetOf(DirectoryEntry, name),
     XtRImmediate, NULL},
{"address", "address", XmtRBuffer,
     XmtSizeOf(DirectoryEntry, address), XtOffsetOf(DirectoryEntry, address),
     XtRImmediate, NULL},
{"comments", "comments", XtRString,
     sizeof(char *), XtOffsetOf(DirectoryEntry, comments),
     XtRImmediate, NULL},
{"type", "type", XtRInt,
     sizeof(int), XtOffsetOf(DirectoryEntry, type),
     XtRImmediate, (XtPointer)PROTOCOL_MODEM},
};

static Widget toplevel;
static Widget directory_dialog, parameters_dialog;
static int current_entry;

static void init_directory(void)
{
    int i;

    /*
     * Normally we'd read our network directory data from a file or something.
     * But here, we're just faking it with some static data.  We've got
     * to malloc space for the big static comment strings, so that they
     * can be freed later.
     */

    for(i=0; i < XtNumber(directory); i++) 
	directory[i].comments = XtNewString(directory[i].comments);
}

static void show_entry(Widget w, XtPointer tag, XtPointer data)
{
    XmListCallbackStruct *list_data = (XmListCallbackStruct *)data;

    current_entry = list_data->item_position - 1;
    XmtDialogSetDialogValues(directory_dialog, &directory[current_entry]);
}


static void do_directory(Widget w, XtPointer tag, XtPointer data)
{
    if (!directory_dialog) {
	Widget list;
	XmString item;
	int i;
	
	directory_dialog = XmtBuildQueryDialog(toplevel, "net_dialog_shell",
					       directory_resources,
					       XtNumber(directory_resources),
					       NULL);

	list = XmtNameToWidget(directory_dialog, "*list");
	for(i = 0; i < XtNumber(directory); i++) {
	    item = XmStringCreateSimple(directory[i].name);
	    XmListAddItem(list, item, 0);
	    XmStringFree(item);
	}
    }
    XmtDialogSetDialogValues(directory_dialog, &directory[current_entry]);
    XtManageChild(directory_dialog);
}

static void edit_directory(Widget w, XtPointer tag, XtPointer data)
{
    /*
     * The comments field of our data structure is an allocated string.
     * We've got to free the old value before we overwrite it with a
     * newly allocated value.
     */
    XtFree(directory[current_entry].comments);
    
    /* Now go get the data from the dialog widgets */
    XmtDialogGetDialogValues(directory_dialog, &directory[current_entry]);
}

static void set_defaults(Widget w, XtPointer tag, XtPointer data)
{
    CommunicationParameters params;

    XmtDialogGetDefaultValues(parameters_dialog, &params, NULL, 0);
    XmtDialogSetDialogValues(parameters_dialog, &params);
}

static void do_parameters(Widget w, XtPointer tag, XtPointer data)
{
    if (!parameters_dialog) {
	parameters_dialog = XmtBuildQueryDialog(toplevel, "params_shell",
						parameter_resources,
						XtNumber(parameter_resources),
						NULL);
    }
    XmtDialogDo(parameters_dialog, &directory[current_entry].params);
}



main(int argc, char **argv)
{
    XtAppContext app;
    
    toplevel = XtAppInitialize (&app, "demo", NULL, 0,
				&argc, argv, NULL, NULL, 0);

    XmtRegisterLayoutParser();
    XmtRegisterWidgetConverter();
    XmtRegisterCallbackConverter();
    XmtRegisterXmStringConverter();
    XmtRegisterPixmapConverter();
    XmtRegisterBitmapConverter();

    XmtRegisterMotifWidgets();
    XmtRegisterXmtWidgets();

    XmtRegisterXtProcedures();
    XmtRegisterXmtProcedures();

    XmtVaRegisterCallbackProcedures(
	      "do_directory", do_directory, XmtRCallbackUnused,
	      "edit_directory", edit_directory, XmtRCallbackUnused,
	      "do_parameters", do_parameters, XmtRCallbackUnused,
	      "show_entry", show_entry, XmtRCallbackUnused,
	      "set_defaults", set_defaults, XmtRCallbackUnused,
	      NULL);

    init_directory();

    XmtCreateChildren(toplevel);

    XtRealizeWidget (toplevel);
    XtAppMainLoop(app);
}
